/* -------------------------------------------------------------------------------------------------------------------------
* Project           :C-to-FPGA Solution for Acclerating Tomographic Reconstruction
* File Name         :uwer_io.c (from project UW Emission Reconstruction) 
*                    Copyright (c) 2004 University of Washington - Imaging Research Laboratory. All rights reserved.
* Version           :1.0 (V1)
* (V1)Author        :Adam Alessio
* (V1)Created       :Wed, Jan 14 2004
* References        :
* Description       :Performs file io
*
******This file may not be distributed or modified without this header*****
*-------------------------------------------------------------------------------------------------------------------------*/


#include "uwer_io.h"

/********************************************************/
/* Function: Sets variables for reconstructed image */
/**/
void prepare_recon_image(struct exec_parameters *params,struct sino_info *sinoInfo,struct recon_image_info *imageInfo)
{
  int frame, plane, jj1,jj2;
  imageInfo->xDim = params->resolution;
  imageInfo->yDim = params->resolution;
  imageInfo->zDim = sinoInfo->nPlanes;

  imageInfo->nFrames = sinoInfo->nFrames;

  imageInfo->xPixelSize = (double)params->tfov/(double)(imageInfo->xDim);
  imageInfo->yPixelSize = (double)params->tfov/(double)(imageInfo->yDim);
  imageInfo->zPixelSize = (double)params->afov/(double)(imageInfo->zDim);

  imageInfo->tfov = params->tfov;
  imageInfo->afov = params->afov;
  imageInfo->nSubSets = params->num_subsets;
  printf("Allocating image size (%d, %d, %d)\n", imageInfo->xDim,imageInfo->yDim,imageInfo->zDim); 
  imageInfo->imageData = (float****)multialloc(sizeof(float),4,imageInfo->nFrames,imageInfo->zDim,imageInfo->yDim,imageInfo->xDim);
  
 for (frame=0;frame<imageInfo->nFrames;frame++)
   for (plane=0;plane<imageInfo->zDim;plane++)
     for(jj1 = 0;jj1<imageInfo->yDim;jj1++)
       for(jj2 = 0;jj2<imageInfo->xDim;jj2++)   
	 imageInfo->imageData[frame][plane][jj1][jj2]=0.0;
  
}

/********************************************************/
/* Function: Sets variables for forward projected data */
/**/
void prepare_forward_projections(struct exec_parameters *params,struct sino_info *sinoInfo,struct recon_image_info *imageInfo)
{
  int frame, plane, angle,proj;
  sinoInfo->nProjs = params->resolution;
  sinoInfo->nAngles = params->resolution;
  sinoInfo->nPlanes = imageInfo->zDim;

  sinoInfo->nFrames=imageInfo->nFrames;

  imageInfo->xPixelSize = (double)params->tfov/(double)(imageInfo->xDim);
  imageInfo->yPixelSize = (double)params->tfov/(double)(imageInfo->yDim);
  imageInfo->zPixelSize = (double)params->afov/(double)(imageInfo->zDim);
  sinoInfo->projSpacing = (double)params->tfov/(double)(sinoInfo->nProjs);
  sinoInfo->angleSpacing = (double)M_PI/(double)sinoInfo->nAngles;
 
  imageInfo->tfov = params->tfov;
  imageInfo->afov = params->afov;
  imageInfo->nSubSets = params->num_subsets;
  printf("Allocating projection size (%d, %d, %d)\n", sinoInfo->nPlanes,sinoInfo->nAngles,sinoInfo->nProjs); 
  sinoInfo->sinogramData = (float****)multialloc(sizeof(float),4,imageInfo->nFrames,sinoInfo->nPlanes,sinoInfo->nAngles,sinoInfo->nProjs);
  
 for (frame=0;frame<imageInfo->nFrames;frame++)
   for (plane=0;plane<sinoInfo->nPlanes;plane++)
     for (angle=0;angle<sinoInfo->nAngles;angle++)
       for (proj = 0; proj<sinoInfo->nProjs;proj++)
	 sinoInfo->sinogramData[0][plane][angle][proj]=0.0;
  
}
  

/********************************************************/
/* Function: Reads sinogram from data file */
/**/
void read_input_sinogram(struct exec_parameters *params,struct sino_info *sinoInfo)
{
  char *exten;
  FILE *fp;
 
    
  // Check Existence of Input File
  if((fp = fopen(params->InputSinogramFileName,"rb"))==NULL){
    Error("Input file does not exist: %s", params->InputSinogramFileName);
  }

  
  
  // Determine type of data file from extension
  if((exten = strrchr(params->InputSinogramFileName,'.'))==NULL)
    Error("Input file does not have extension: %s", params->InputSinogramFileName);
  else{
    if(!(strstr(exten,"fld")==NULL)){
      read_input_avs_sinogram(sinoInfo,fp);
     
    }
    else if(!(strstr(exten,"S")==NULL)){
      read_input_ecat_sinogram(sinoInfo,fp, params);
    }
    else
      Error("Input file type '%s' not supported",exten);
  }

  fclose(fp);

  // Assign other values now that have sinogram size
  sinoInfo->angleSpacing = (double)M_PI/(double)sinoInfo->nAngles;
  sinoInfo->projSpacing = (double)params->tfov/(double)(sinoInfo->nProjs);  // AMA nProjs change Since have finite width projections
  sinoInfo->projWidth =  sinoInfo->projSpacing * params->projWidthMultiple;

  if(params->ReconPlanes[0] == -1){  // If ReconPlanes not set at command line
    params->ReconPlanes[0] = 0;
    params->ReconPlanes[1] = sinoInfo->nPlanes-1; // -1 because inclusive
  }

   /***-- Estimate Variance of Data for PWLS --***/
  /* Similar to method from Fessler '94 */
  
  if(params->reconType==4)
    {
      float sum, varG,g, kernelSmooth[7];
      int frame, plane, angle, proj, i;
      printf("Estimating Variance of Data for PWLS\n");
      /* Smooth with 1D Gaussian */
      varG = 1.0/(2*sqrt(2*log(2)));  // Standard Deviations
      varG = varG*varG;
      for(i = 0;i<7;i++)
	kernelSmooth[i] = exp(- ((i-3)*(i-3))/(2*varG));
      sum = 0.0;
      for(i = 0;i<7;i++)
	sum +=kernelSmooth[i];
      for(i = 0;i<7;i++)
	kernelSmooth[i] = kernelSmooth[i]/sum; 
    
    sinoInfo->VarSinogramData = (float****)multialloc(sizeof(float),4,sinoInfo->nFrames,sinoInfo->nPlanes,sinoInfo->nAngles,sinoInfo->nProjs);

    for(frame=0;frame<sinoInfo->nFrames;frame++)
      for (plane=0;plane<sinoInfo->nPlanes;plane++)
	for (angle=0;angle<sinoInfo->nAngles;angle++)
	  for (proj = 0; proj<sinoInfo->nProjs;proj++){
	    sinoInfo->VarSinogramData[frame][plane][angle][proj] = 0.0;
	    for(i = -3;i<=3;i++){
	    
	      if(((proj+i)>=0) && ((proj+i)<sinoInfo->nProjs))
		{
		  /*  if(params->AW_flag)
		    g =  sinoInfo->AWsinogramData[plane][angle][proj+i];
		    else */
		  g = 1.0;
		  sinoInfo->VarSinogramData[frame][plane][angle][proj] += sinoInfo->sinogramData[frame][plane][angle][proj+i] * g;
		}
	      if(sinoInfo->VarSinogramData[frame][plane][angle][proj]<7.0)
		sinoInfo->VarSinogramData[frame][plane][angle][proj] = 7.0;
	    }
	  }
    //#define ONEVARSINO 
    #ifdef OUTPUT_ONEVARSINO 
    {
      FILE *f2;
    int proj,angle;
    
    f2 = fopen("sinoVarSlice.yes","w");
 
    for (angle=0;angle<sinoInfo->nAngles;angle++){
	for (proj=0;proj<sinoInfo->nProjs;proj++){
	fprintf(f2,"%f   ", sinoInfo->VarSinogramData[0][35][angle][proj]);
      }
      fprintf(f2,"\n");
    }
    fclose(f2);
    
    f2 = fopen("sinoSlice.yes","w");
 
    for (angle=0;angle<sinoInfo->nAngles;angle++){
	for (proj=0;proj<sinoInfo->nProjs;proj++){
	fprintf(f2,"%f   ", sinoInfo->sinogramData[0][35][angle][proj]);
      }
      fprintf(f2,"\n");
    }
    fclose(f2);
    //  exit(0);
  
    }
#endif  

    }

	    
  
  // TEST Sino Read
#ifdef OUTPUT_ONESINO 
  {
    FILE *f2;
    int proj,angle;
    
    f2 = fopen("sinoSlice.yes","w");
 
    for (angle=0;angle<sinoInfo->nAngles;angle++){
	for (proj=0;proj<sinoInfo->nProjs;proj++){
	fprintf(f2,"%f   ", sinoInfo->sinogramData[0][85][angle][proj]);
      }
      fprintf(f2,"\n");
    }
    fclose(f2);
    exit(0);
  }
#endif  
 

  }

/********************************************************/
/* Function: Applys attenuation correction to the data (prior to recon) */
/**/
void apply_attenuation_correction(struct exec_parameters *params,struct sino_info *sinoInfo)
{
  float *tempFloat;
  int cnts, plane, angle,proj, index;
  FILE *fp;
  
  if(!params->attenCorrect_flag)
    printf("Not applying attenuation correction\n");
  else
    {
      
      printf("\n------------\nNotice:::\\nApplying Attenuation Correction\n");
      if((fp = fopen(params->AttenCorrectFileName,"r"))==NULL){
	printf("\nCan not read attenuation input file %sn\n\a\n", params->AttenCorrectFileName);
	exit(0);
      }
      
      tempFloat = (float*)get_spc(sinoInfo->nAngles*sinoInfo->nProjs,sizeof(float)); 
     
      cnts = fread(tempFloat,sizeof(float), sinoInfo->nAngles*sinoInfo->nProjs,fp);
      fclose(fp);
     
      for (plane=0;plane<sinoInfo->nPlanes;plane++){
	index = 0;
	for (angle=0;angle<sinoInfo->nAngles;angle++)
	  for (proj = 0; proj<sinoInfo->nProjs;proj++)
	    sinoInfo->sinogramData[0][plane][angle][proj] =sinoInfo->sinogramData[0][plane][angle][proj]*(float)tempFloat[index++];
     }
     
     fp = fopen("attenTest.dd","w"); 
     index = 0;
     for (angle=0;angle<sinoInfo->nAngles;angle++){
       for (proj = 0; proj<sinoInfo->nProjs;proj++){
	 
     	 fprintf(fp,"%f   ", tempFloat[index++]); 
       } 
       fprintf(fp,"\n");
     }
     fclose(fp); 
     
     free(tempFloat);
    }
}

/********************************************************/
/* Function: Reads sinogram from AVS format data file */
/**/
      
void read_input_avs_sinogram(struct sino_info *sinoInfo,FILE *fp)
{
  int ncid;
  char *st2, st1[100];
  int st3,cnts,frame,plane,angle,proj,index,dimid,little_endian;
  float *tempFloat;
  
  printf("Reading AVS File Format\n");
  for (;;)
    {
      ncid = fscanf(fp, "%s", st1);
      if(!(strstr(st1,"ndim=")==NULL)){
	st2 = strstr(st1,"=");
	dimid = (int)atol(st2+1);
	if(dimid != 3){
	  Error("Wrong number of dimensions in input file!! %d\n", dimid);
	} 
      }
      else if(!(strstr(st1,"dim1=")==NULL)){
	st2 = strstr(st1,"=");
	sinoInfo->nProjs = (int)atol(st2+1);
      }
       else if(!(strstr(st1,"dim2=")==NULL)){
	st2 = strstr(st1,"=");
	sinoInfo->nAngles = (int)atol(st2+1);
      }
       else if(!(strstr(st1,"dim3=")==NULL)){
	st2 = strstr(st1,"=");
	sinoInfo->nPlanes = (int)atol(st2+1);
	printf("Sinogram Size = (%d,%d,%d)\n", sinoInfo->nProjs, sinoInfo->nAngles, sinoInfo->nPlanes);
       }
       else if(!(strstr(st1,"data=")==NULL)){
	 if(!(strstr(st1,"xdr_float")==NULL))
	   little_endian = 0;
	 else
	   little_endian = 1;
	 break;
       }
	

    }
      

  for(;;){
    st3 =  getc(fp);
    if(st3==12){
      st3 =  fgetc(fp);
      if(st3==12)
	{
	  //	  printf("found double feed\n");
	  break;break;
	}
      else
	{
	    Error("Input file does not contain double form feed!!\nExiting\n");
	  }
    }
  }

  sinoInfo->nFrames = 1;
  
  tempFloat = (float*)get_spc(sinoInfo->nPlanes*sinoInfo->nAngles*sinoInfo->nProjs,sizeof(float));   
  cnts = fread(tempFloat,sizeof(float), sinoInfo->nPlanes*sinoInfo->nAngles*sinoInfo->nProjs,fp);

  
  
  ncid = 1;
  if((little_endian) && (!(*(char*)&ncid))){
    printf("Big endian machine, reading little endian data-> swap\n");
    for(index = 0;index<cnts;index++)
      SWAP_FLOAT(tempFloat[index]);
  }
  if(!(little_endian) && ((*(char*)&ncid))){
    printf("Little endian machine, reading big endian data->swap\n");
    for(index = 0;index<cnts;index++)
      SWAP_FLOAT(tempFloat[index]);
  }  

sinoInfo->sinogramData = (float****)multialloc(sizeof(float),4,sinoInfo->nFrames,sinoInfo->nPlanes,sinoInfo->nAngles,sinoInfo->nProjs);
 index = 0;
  for(frame=0;frame<sinoInfo->nFrames;frame++)
    for (plane=0;plane<sinoInfo->nPlanes;plane++)
      for (angle=0;angle<sinoInfo->nAngles;angle++)
	for (proj = 0; proj<sinoInfo->nProjs;proj++){
	  //  tempFloat[index]*=1000.0;
	  if(tempFloat[index]<0.0)
	    tempFloat[index] = 0.0;
	  //	    printf("Negative Sinogram Entries: (%d, %d, %d, %d) = %f\n", frame, plane, angle, proj, tempFloat[index]);
	  sinoInfo->sinogramData[frame][plane][angle][proj] =(float)tempFloat[index++];  //AMA TESTING REMOVE +TOL LATER!!!!
	}
  //  printf("ADDED constant to all data to keep >0, change later \n");
  
  free((void*)tempFloat);

  {
    double sum;
    sum = 0.0;
   for(frame=0;frame<sinoInfo->nFrames;frame++)
    for (plane=0;plane<sinoInfo->nPlanes;plane++)
      for (angle=0;angle<sinoInfo->nAngles;angle++)
	for (proj = 0; proj<sinoInfo->nProjs;proj++)
	  sum += sinoInfo->sinogramData[frame][plane][angle][proj];
  printf("sum of input sinogram %f\n", sum);
  }
  }  // END read_input_avs_sinogram()



/********************************************************/
/* Function: Reads image from AVS format data file */
/**/
      
void read_input_avs_image(struct recon_image_info *imageInfo,struct exec_parameters *params)
{
  FILE *fp;
  int ncid;
  char *st2, st1[100];
  int st3,cnts,frame,plane,nX,nY,x,y,nplanes,index,dimid,little_endian;
  float *tempFloat;

  if((fp=fopen(params->OutputImageFileName,"r")) == NULL)
    Error("Can not open input image file:%s\n", params->AttenCorrectFileName);
    
  printf("Reading AVS File Format- AW File\n");
  for (;;)
    {
      ncid = fscanf(fp, "%s", st1);
      if(!(strstr(st1,"ndim=")==NULL)){
	st2 = strstr(st1,"=");
	dimid = (int)atol(st2+1);
	if(dimid != 3){
	  Error("Wrong number of dimensions in input file!! %d\n", dimid);
	} 
      }
      else if(!(strstr(st1,"dim1=")==NULL)){
	st2 = strstr(st1,"=");
	nX = (int)atol(st2+1);
      }
       else if(!(strstr(st1,"dim2=")==NULL)){
	st2 = strstr(st1,"=");
	nY = (int)atol(st2+1);
      }
       else if(!(strstr(st1,"dim3=")==NULL)){
	st2 = strstr(st1,"=");
	nplanes = (int)atol(st2+1);
	printf("Image Size = (%d,%d,%d)\n", nplanes,nY, nX);
       }
       else if(!(strstr(st1,"data=")==NULL)){
	 if(!(strstr(st1,"xdr_float")==NULL))
	   little_endian = 0;
	 else
	   little_endian = 1;
	 break;
       }
	

    }
     
  
  for(;;){
    st3 =  getc(fp);
    if(st3==12){
      st3 =  fgetc(fp);
      if(st3==12)
	{
	  //	  printf("found double feed\n");
	  break;break;
	}
      else
	{
	    Error("Input file does not contain double form feed!!\nExiting\n");
	  }
    }
  }

  imageInfo->nFrames = 1;
  imageInfo->zDim = nplanes;
  imageInfo->yDim = nY;
  imageInfo->xDim = nX;
  
  
  tempFloat = (float*)get_spc(nX*nY*nplanes,sizeof(float));   
  cnts = fread(tempFloat,sizeof(float), nX*nY*nplanes,fp);

  ncid = 1;
  if((little_endian) && (!(*(char*)&ncid))){
    printf("Big endian machine, reading little endian data-> swap\n");
    for(index = 0;index<cnts;index++)
      SWAP_FLOAT(tempFloat[index]);
  }
  if(!(little_endian) && ((*(char*)&ncid))){
    printf("Little endian machine, reading big endian data->swap\n");
    for(index = 0;index<cnts;index++)
      SWAP_FLOAT(tempFloat[index]);
  }  

  imageInfo->imageData = (float****)multialloc(sizeof(float),4,imageInfo->nFrames,nplanes,nY,nX);
  index = 0;
 
    for (plane=0;plane<nplanes;plane++)
      for (y=0;y<nY;y++)
	for (x = 0;x<nX;x++){
	  imageInfo->imageData[0][plane][y][x] =(float)tempFloat[index++];
	}
  
  free((void*)tempFloat);

}  // END read_input_avs_image()

  
/********************************************************/
/* Function: Parses command line or configuration file to set the reconstruction parameters */
/**/
void parse_cmdline_configfile(struct exec_parameters *params,int argc,char **argv)
{
  int i,stat;  
  FILE *fp, *fptest;   
  char st1[256];
  int slen;
  short   configFile_flag, requiredItemCount;

 
	params->num_iterations = 5;
	params->num_subsets=1;
	
	params->interpolate_flag=0;
	params->resolution=128;
	configFile_flag = 0;
	params->reconType  = 1;
	requiredItemCount = 0;
	params->ReconPlanes[0] = -1;  // Flag for seeing if assigned later
	params->ReconPlanes[1] = -1;  // Flag for seeing if assigned later
	params->tfov = 500.0;
	params->afov = 10.0;  // NOT USED FOR ANYTHING CURRENTLY
	params->sliceThick=4.222;  // Slice Thickness (mm)
	params->fbp_type = 1;    // filter type 1=Hamming, 0 = ramp
	params->fbp_cutoff = 1.0;  // Cutoff frequency for fbp recon
	params->recon_mask_flag = 0;
	params->attenCorrect_flag = 0;
	params->projWidthMultiple = 1.0;  // Will multiply by projSpacing in read_input_sinogram()
	params->anatomical_labels_flag = 0;
	params->p = 2.0;
	params->globalBeta = 1.0;
	params->AW_flag = 0;
	params->OutEachIt = 0;
	params->penalty_type = (short)0;   // Index for type of penalty term 1 for labels, 0 for quadratic
	
	params->InitialImageFileName = (char*)get_spc(5,sizeof(char));
	sprintf(params->InitialImageFileName,"none");

	/*--------- Read Parameters from the Command Line -----*/
	/* If -C set, will read parameters exclusively from configuration file */
	
	for (;;)
    { 
		i= getopt(argc, argv, "i:t:c:L:b:u:q:z:s:mow:I:f:r:p:R:S:W:A:C:O:P:");
		if (i==EOF)
		  {
			break;;
		  }
			else
        {
		
			if(i=='t')
			  {
			    params->reconType = atoi(optarg);
			    if(params->reconType>5)
			      usage_error();
			  }
			
			else if(i=='c'){
			  params->fbp_cutoff = atof(optarg);
			}
			else if(i=='R'){
			  params->fbp_type = atoi(optarg);
			}

			else if (i=='r')
			{
				params->resolution=atoi(optarg);
			}
			
			else if(i=='S')
			  {
			    requiredItemCount++;
			    slen = (int)strlen(optarg)+1;
			    params->InputSinogramFileName = (char*)get_spc(slen,sizeof(char));
			    strcpy(params->InputSinogramFileName,optarg);
			}
			
			else if (i=='O')
			  {
			    requiredItemCount++;
			    slen = (int)strlen(optarg)+1;  // Had some odd bus error with this when >64 -> Output file location did not exist
			    params->OutputImageFileName = (char*)get_spc(slen,sizeof(char));
			    strcpy(params->OutputImageFileName,optarg);
			   
			  }
			else if (i == '?')
			  usage_error();
			
			
		}
    }
	if((params->reconType ==4) &&(params->AW_flag)){
	  printf("Attempting PWLS with AW - not necessary, turning off AW\n");
	  params->AW_flag = 0;
	}
	    
	  
	
	if(requiredItemCount<2)
	  usage_error();
	
	/*-------- END of reading parameters ----------*/

	
	OpenLog(params->OutputImageFileName);  // Open Log File   
	
	if(configFile_flag){
		printf("Configuration file : %sn",params->config_fname);	
	}

	printf("Settings:\n     input file:%s\n     output file:%s\n",params->InputSinogramFileName,params->OutputImageFileName);
	printf("     reconstruct planes [%d, %d]\n", params->ReconPlanes[0],params->ReconPlanes[1]);
	printf("     scanner TFOV: %f\n     output resolution: %d\n", params->tfov, params->resolution);
	printf("     slice thickness: %f\n",params->sliceThick);

	if (params->reconType==3){
	  printf("   FBP filter = %d, FBP cutoff frequency = %f\n", params->fbp_type, params->fbp_cutoff);
	}
	else{
	  printf("     Initial Image: %s\n",params->InitialImageFileName); 
	  if(params->reconType==1 || params->reconType==2 || params->reconType==5)
	    printf("     num_iterations: %d\n     num_subsets: %d\n", params->num_iterations, params->num_subsets);
	  else if(params->reconType==4)
	    printf("     PWLS num_iterations: %d\n     prior shape = %f, prior strength = %f, penalty type=%d\n", params->num_iterations, params->p, params->globalBeta, params->penalty_type);
	  else
	    usage_error();
	}
	//	printf("     reconstruct planes: %d  through %d\n", params->ReconPlanes[0], params->ReconPlanes[1]);		
	if(params->OutEachIt)
	  printf("   **Writing out image at each iteration**\n");

	  
	fflush(stderr);
	fflush(stdout);
	
}  // End parse_cmdline_configfile()


/********************************************************/
/* Function: Displays error message for wrong execution settings */
/**/

void usage_error()
{
	printf("Usage:\n");
	printf("uwer [-t recon_type(1=OSEM, 2=OSEM+PSF2+1, 3=FBP, 4=PWLS+Prior+PSF21)] \n     [-f field_of_view_mm (550)] [-r resolution (128)] \n     [-p recon_planes(0 #planes)] [-m apply recon mask] [-C config_file(none)] [-z Slick Thickness (mm)] \n     -S input_sinogram_file [-L input_anatomical_label_file] \n     -O output_file_name [-P PSF_file_name]\n\n");

	printf("Statistical Recon Settings: [-i #iter(5)] [-s #subsets(4)] \n     [-w projection_width_multiple (1.0)] [-b prior_beta(1.0)] \n     [-q prior_shape(2.0)]  [-I initial_image(none)] [-u penalty type(0 quadratic, 1 labels)]\n\n");
	printf("FBP Recon Settings: [-R filter_type 0=ramp, 1=Hanning (1)][-c cut_off_frequency (1.0)]\n\n");
	printf("Example: \n uwer -t 3 -f 300 -R 0 -c 0.8 -S mySinogram.fld -O myReconImage.fld\n (Runs FBP reconstruction with 80\% cutoff ramp filter)\n\n");

	exit(1);
}

/********************************************************/
/* Function: outputs final reconstructed image  */
/**/

void output_recon_image(struct exec_parameters *params,struct recon_image_info *imageInfo)
{
  FILE *fp;
  int frame,plane,y,x,index;
  float *temp_array;
  char tmp_str[256], *reconInfo;


  // Writing reconstruction information into comment
  reconInfo = (char*)get_spc(1024,sizeof(char));
  sprintf(reconInfo,"[INPUT_SINO] = %s\n", params->InputSinogramFileName);

  sprintf(tmp_str,"[TFOV] = %f\n[RES] = %d\n[RECON_PLANES] = [%d, %d]\n",  params->tfov,params->resolution, params->ReconPlanes[0],params->ReconPlanes[1]);
  strcat(reconInfo, tmp_str);
  // sprintf(tmp_str,"[ATTEN_CORRECT] = %d\n", params->attenCorrect_flag);

   if(params->reconType==5)
      sprintf(tmp_str,"[METHOD] = OSEM w/ multiresolution image\n");
   
  // OSEM INFO
  if(params->reconType<3){
    if(params->reconType==1)
      sprintf(tmp_str,"[METHOD] = OSEM\n");
    if(params->reconType==2)
      sprintf(tmp_str,"[METHOD] = OSEM with 2+1 SRF: %s \n", params->PSFfileName);
    
    strcat(reconInfo, tmp_str);
    sprintf(tmp_str,"[ITERATIONS] = %d\n[NUMSUBSETS] = %d\n",params->num_iterations, params->num_subsets);
    strcat(reconInfo, tmp_str);
  }
  
  // PWLS Info
  if((params->reconType==4))
    {
      sprintf(tmp_str,"[METHOD] = PWLS with 2+1 SRF\n");
      strcat(reconInfo, tmp_str);
      sprintf(tmp_str,"[ITERATIONS] = %d\n", params->num_iterations);
      strcat(reconInfo, tmp_str);
      sprintf(tmp_str,"[PRIOR SHAPE] = %2.2f\n[PRIOR GLOBAL BETA] = %f\n", params->p, params->globalBeta);
      strcat(reconInfo, tmp_str);
      sprintf(tmp_str,"[PSF File] = %s\n", params->PSFfileName);
      if(params->anatomical_labels_flag){
	sprintf(tmp_str,"[Anatomical Labels] = %s\n", params->AnatomicalLabelName);
	strcat(reconInfo, tmp_str);
      }
    }
      
  // PWLS and OSEM Info
  if(params->reconType !=3)
    {
    sprintf(tmp_str,"[PROJ_WIDTH_MULTIPLE] = %f\n",params->projWidthMultiple);
    strcat(reconInfo, tmp_str);
    sprintf(tmp_str,"[INITIAL_IMAGE] = %s\n",params->InitialImageFileName);
    strcat(reconInfo, tmp_str);
    if(params->penalty_type==0)
      sprintf(tmp_str,"[PENALTY_TYPE] = quadratic\n");
    else if(params->penalty_type==1)
      sprintf(tmp_str,"[PENALTY_TYPE] = labels\n");
    strcat(reconInfo, tmp_str);
    
    }
  
  
  else if(params->reconType==3){ // FBP Info
    sprintf(tmp_str,"[METHOD] = FBP\n");
    
    strcat(reconInfo, tmp_str);
    sprintf(tmp_str,"[FILTER] = %d\n[CUTOFF] = %f\n", params->fbp_type, params->fbp_cutoff);
    strcat(reconInfo, tmp_str);
  }
  
  if(params->AW_flag){    //  (AW) Info
    sprintf(tmp_str,"[ATTEN WEIGHT FILE] = %s\n", params->AttenCorrectFileName);;
    strcat(reconInfo, tmp_str);
  }
    
#ifdef TEST_OUTPUTIMAGE
  // TEST image write
  {
    FILE *f2;
    
    f2 = fopen("image.yes","w");
    for (y=0;y<imageInfo->yDim;y++){
        for (x=0;x<imageInfo->xDim;x++)
	  fprintf(f2,"%f   ", imageInfo->imageData[0][85][y][x]);
      fprintf(f2,"\n");
    }
    fclose(f2);
  }
#endif
  
  if((fp=fopen(params->OutputImageFileName,"wb")) == NULL)
    Error("\nCan't open ouput file %s\n\a\n",params->OutputImageFileName);

  printf("Writing reconstructed image to file: %s\n", params->OutputImageFileName);
  
  temp_array = (float*)get_spc(imageInfo->xDim*imageInfo->yDim*imageInfo->zDim*imageInfo->nFrames,sizeof(float));
  
  index = 0;
  for (frame=0;frame<imageInfo->nFrames;frame++)
    for (plane=0;plane<imageInfo->zDim;plane++) 
      for (x=imageInfo->xDim-1;x>=0;x-=1)   // To Have spine down
	  for (y=0;y<imageInfo->yDim;y++)
	  temp_array[index++] = (float)imageInfo->imageData[frame][plane][y][x];

          
// Place Header Information in File
  fprintf(fp, "# AVS field file        (avs_write_us)\n");
  
  fprintf(fp,"# %s",reconInfo);
  fprintf(fp,"ndim=3\n");
  fprintf(fp,"dim1=%d\n",imageInfo->xDim);
  fprintf(fp,"dim2=%d\n",imageInfo->yDim);
  fprintf(fp,"dim3=%d\n",imageInfo->zDim);
  fprintf(fp,"dim1Size=%f\n", params->tfov/(float)imageInfo->xDim);
  fprintf(fp,"dim2Size=%f\n", params->tfov/(float)imageInfo->yDim);
  fprintf(fp,"dim3Size=%f\n", params->sliceThick);
  fprintf(fp,"nspace=3\n");
  fprintf(fp, "veclen=%d\n",	1);
  x = 1;
  if (*(char*)&x)  // If this is a little endian machine
    fprintf(fp,"data=float_le\n");
  else
    fprintf(fp,"data=xdr_float\n");
  fprintf(fp, "field=%s\n",	"uniform");  
  fprintf(fp, "\f\f");

  fwrite(temp_array, sizeof(float), imageInfo->xDim*imageInfo->yDim*imageInfo->zDim*imageInfo->nFrames, fp);
  free(temp_array);
  fclose(fp);
  
  fflush(stderr);
  fflush(stdout);
} // End output_recon_image()


/********************************************************/
/* Function: outputs an intermediate image  */
/**/

void output_an_image(char *fnm, struct exec_parameters *params,struct recon_image_info *imageInfo)
{
  FILE *fp;
  int frame,plane,y,x,index;
  float *temp_array;
  

 if((fp=fopen(fnm,"wb")) == NULL)
    Error("\nCan't open ouput file %s\n\a\n",fnm);

  printf("\tWriting intermediate reconstructed image to file: %s\n", fnm);
  
  temp_array = (float*)get_spc(imageInfo->xDim*imageInfo->yDim*imageInfo->zDim*imageInfo->nFrames,sizeof(float));
  
  index = 0;
  for (frame=0;frame<imageInfo->nFrames;frame++)
    for (plane=0;plane<imageInfo->zDim;plane++) 
      for (x=imageInfo->xDim-1;x>=0;x-=1)   // To Have spine down
	  for (y=0;y<imageInfo->yDim;y++)
	  temp_array[index++] = (float)imageInfo->imageData[frame][plane][y][x];
  
  // Place Header Information in File
  fprintf(fp, "# AVS field file        (avs_write_us)\n");
  
  fprintf(fp,"ndim=3\n");
  fprintf(fp,"dim1=%d\n",imageInfo->xDim);
  fprintf(fp,"dim2=%d\n",imageInfo->yDim);
  fprintf(fp,"dim3=%d\n",imageInfo->zDim);
  fprintf(fp,"dim1Size=%f\n", 1000.0*params->tfov/(float)imageInfo->xDim);
  fprintf(fp,"dim2Size=%f\n", 1000.0*params->tfov/(float)imageInfo->yDim);
  fprintf(fp,"dim3Size=%f\n", 4.2222);
  fprintf(fp,"nspace=3\n");
  fprintf(fp, "veclen=%d\n",	1);
  x = 1;
  if (*(char*)&x)  // If this is a little endian machine
    fprintf(fp,"data=float_le\n");
  else
    fprintf(fp,"data=xdr_float\n");
  fprintf(fp, "field=%s\n",	"uniform");  
  fprintf(fp, "\f\f");

  fwrite(temp_array, sizeof(float), imageInfo->xDim*imageInfo->yDim*imageInfo->zDim*imageInfo->nFrames, fp);
  free(temp_array);
  fclose(fp);
  
  fflush(stderr);
  fflush(stdout);
} // End output_an_image()

/********************************************************/
/* Function: outputs an intermediate image  */
/**/

void output_sinogram(char *fnm, struct exec_parameters *params,struct sino_info *sinoInfo)
{
  FILE *fp;
  int frame,plane,angle,proj,index,x;
  float *temp_array;
  

 if((fp=fopen(fnm,"wb")) == NULL)
    Error("\nCan't open ouput file %s\n\a\n",fnm);

  printf("\tWriting forward projected data to file: %s\n", fnm);
  
  temp_array = (float*)get_spc(sinoInfo->nPlanes*sinoInfo->nAngles*sinoInfo->nProjs*sinoInfo->nFrames,sizeof(float));

  index = 0;
  for (frame=0;frame<sinoInfo->nFrames;frame++)
    for (plane=0;plane<sinoInfo->nPlanes;plane++)
      for (angle=0;angle<sinoInfo->nAngles;angle++)
		for (proj = 0; proj<sinoInfo->nProjs;proj++)	
			temp_array[index++] = (float)sinoInfo->sinogramData[frame][plane][angle][proj];
 
  // Place Header Information in File
  fprintf(fp, "# AVS field file        (avs_write_us)\n");
  fprintf(fp,"# Simulated Forward Projections\n");
  fprintf(fp,"ndim=3\n");
  fprintf(fp,"dim1=%d\n",sinoInfo->nProjs);
  fprintf(fp,"dim2=%d\n",sinoInfo->nAngles);
  fprintf(fp,"dim3=%d\n",sinoInfo->nPlanes);
  fprintf(fp,"dim1Size=%f\n", 1.0);
  fprintf(fp,"dim2Size=%f\n", 1.0);
  fprintf(fp,"dim3Size=%f\n", 1.0);
  fprintf(fp,"nspace=3\n");
  fprintf(fp, "veclen=%d\n",	1);
  x = 1;
    
if (*(char*)&x)  // If this is a little endian machine
    fprintf(fp,"data=float_le\n");
  else
  fprintf(fp,"data=xdr_float\n");
  fprintf(fp, "field=%s\n",	"uniform");  
  fprintf(fp, "\f\f");

  fwrite(temp_array, sizeof(float),sinoInfo->nPlanes*sinoInfo->nAngles*sinoInfo->nProjs*sinoInfo->nFrames, fp);
  free(temp_array);
  fclose(fp);
  
  fflush(stderr);
  fflush(stdout);
} // End output_an_image()

/********************************************************/
/* Function: Loads system response information from file  */
/**/
void load_system_response_function(struct system_matrix_info *systemInfo,struct exec_parameters *params, struct sino_info *sinoInfo)
{
  char st1[50],*st2;
  FILE *fp;
  int variantS, numS, numZ,cnts,vs,ns,nz,index,st3;
  float *tempFloat;


 
  /*
  printf("Reading psf from psfTest_delta.psf\n");
  if((fp = fopen("psfTest_delta.psf","r"))==NULL){
    Error("System response file does not exist");
    }
  */
  /*
  printf("Reading psf from   psfTest_LeePoints4.psf\n");
  if((fp = fopen("psfTest_LeePoints4.psf","r"))==NULL){
    Error("System response file does not exist");
  }
  */
  /*
    Generated from single extrapolation to 2+1 Function
    printf("Reading psf from psf21_advance_HALFf.psf\n");
    if((fp = fopen("psf21_advance_HALFf.psf","r"))==NULL){
    Error("System response file does not exist");
    }
  */
     // Generated from double extrapolation in z and s to form 2+2 function
//     printf("Reading psf from psf21_advance_HALF_reduced.psf\n");
//     if((fp = fopen("psf21_advance_HALF_reduced.psf","r"))==NULL){
//     Error("System response file does not exist");
//     }
  systemInfo->systemResponse_flag = 1;
  printf("Reading System Response:%s\n", params->PSFfileName);
  if((fp = fopen(params->PSFfileName,"r"))==NULL){


    printf("System response file does not exist\n");
    printf("Using Unity for response -  No added system information\n");
    variantS = sinoInfo->nProjs;
    numS = 1;
    numZ = 1;
    tempFloat = (float*)get_spc(variantS*numS*numZ,sizeof(float));   

    systemInfo->srf = (float***)multialloc(sizeof(float),3,variantS,numZ,numS);
    systemInfo->srfDim[0] = variantS;
    systemInfo->srfDim[1] = numZ;
    systemInfo->srfDim[2] = numS;
    for(vs = 0;vs<variantS;vs++)
      systemInfo->srf[vs][0][0] = 1.0;
    
  }
  
 
  else{
  // Using AVS style file
    for (;;)
    {
      index = fscanf(fp, "%s", st1);
      if(!(strstr(st1,"ndim=")==NULL)){
	st2 = strstr(st1,"=");
	nz = (int)atol(st2+1);
	if(nz != 3){
	  Error("Wrong number of dimensions in input file!! %d\n", nz);
	} 
      }
      else if(!(strstr(st1,"zCenter=")==NULL)){  // For later AMA
	st2 = strstr(st1,"=");
	variantS = (int)atol(st2+1);
      }
      else if(!(strstr(st1,"variantS=")==NULL)){
	st2 = strstr(st1,"=");
	variantS = (int)atol(st2+1);
      }
       else if(!(strstr(st1,"numZ=")==NULL)){
	st2 = strstr(st1,"=");
	numZ = (int)atol(st2+1);
      }
       else if(!(strstr(st1,"numS=")==NULL)){
	st2 = strstr(st1,"=");
	numS = (int)atol(st2+1);
	printf("System Response Size = (%d,%d,%d)\n",variantS, numZ, numS);
	break;
       }
      /* Don't need this functionality */
      /*   else if(!(strstr(st1,"data=")==NULL)){
	 if(!(strstr(st1,"xdr_float")==NULL))
	 little_endian = 0;
	 else
	 little_endian = 1;
	 break;
	 }
      */
    }   

  for(;;){
    st3 =  getc(fp);
    if(st3==12){
      st3 =  fgetc(fp);
      if(st3==12)
	{
	  break;break;
	}
      else
	{
	    Error("Input file does not contain double form feed!!\nExiting\n");
	  }
    }
  }

  if(variantS != sinoInfo->nProjs)
    Error("System Reponse File does not have correct projections dimension for this sinogram\n");

  
  tempFloat = (float*)get_spc(variantS*numS*numZ,sizeof(float));   
  cnts = fread(tempFloat,sizeof(float), variantS*numS*numZ,fp);

  systemInfo->srf = (float***)multialloc(sizeof(float),3,variantS,numZ,numS);
  systemInfo->srfDim[0] = variantS;
  systemInfo->srfDim[1] = numZ;
  systemInfo->srfDim[2] = numS;

  
  index = 0;
  for(vs = 0;vs<variantS;vs++)
    for(nz = 0;nz<numZ;nz++)
      for(ns = 0;ns<numS;ns++)
	systemInfo->srf[vs][nz][ns] = tempFloat[index++];

  // Switch occuring between 71 and 72, change to 70 71
  /* for(vs = variantS/2;vs<vs;vs++) */
/*     for(nz = 0;nz<numZ;nz++) */
/*       for(ns = 0;ns<numS;ns++) */
/* 	systemInfo->srf[vs][nz][ns] =systemInfo->srf[vs-variantS/2][nz][ns];  */
 /*  for(vs = 0;vs<variantS;vs++)  */
/*     { */
/*       systemInfo->srf[vs][2][4] = .25; */
/*       systemInfo->srf[vs][2][3] = .094; */
/*       systemInfo->srf[vs][2][5] = .094; */
/*       systemInfo->srf[vs][2][2] = .094; */
/*       systemInfo->srf[vs][2][6] = .094; */
/*       systemInfo->srf[vs][2][1] = .094; */
/*       systemInfo->srf[vs][2][7] = .094; */
/*     } */

      
/*       systemInfo->srf[vs][1][4] = .094; */
/*       systemInfo->srf[vs][1][3] = .094; */
/*       systemInfo->srf[vs][1][5] = .094; */
/*       systemInfo->srf[vs][3][4] = .094; */
/*       systemInfo->srf[vs][3][3] = .094; */
/*       systemInfo->srf[vs][3][5] = .094; */
/*     } */
  
  free(tempFloat);
  }
  if(0)
  {
    vs = 100;
    FILE *f2;
    
    f2 = fopen("testSRF.yes","w");

    for(nz = 0;nz<numZ;nz++){
      for(ns = 0;ns<numS;ns++)
	fprintf(f2,"%f   ",systemInfo->srf[vs][nz][ns]);
      fprintf(f2,"\n");
    }
    fclose(f2);

  }


  
  fflush(stderr);
  fflush(stdout);
}
